import geopandas as gpd
import numpy as np
import pandas as pd

from geokde._kernels import (
    epanechnikov_raw,
    epanechnikov_scaled,
    quartic_raw,
    quartic_scaled,
    triweight_raw,
    triweight_scaled,
)


def validate_transform(
    arg: int | float | str,
    arg_name: str,
    points: gpd.GeoDataFrame | gpd.GeoSeries,
) -> np.ndarray:
    """Validation and transformation function for weight and radius arguments.

    If `arg` is present in the GeoDataFrame's columns, said column will be used
    preferentially and converted to ndarray, assuming it is complete and of numeric
    dtype.
    If `arg` is not present in the GeoDataFrame's columns but is int or float, an
    ndarray of equal length to the GeoDataFrame will be populated with the given value.

    Parameters
    ----------
    arg : int | float | str
        Value of the argument to validate and transform.
    arg_name : str
        Name of the argument to validate and transform.
    points : geopandas.GeoDataFrame | geopandas.GeoSeries
        A GeoDataFrame or GeoSeries of point geometries to validate against.

    Returns
    -------
    arg : numpy.ndarray
        ndarray-transformation of `arg`.

    Raises
    ------
    ValueError
        If the column in which weight values are present is not complete or the value
        for `arg` is not appropriate.
    TypeError
        If the column in which weight values are present is not of a numeric dtype.
    """
    if arg in points.columns:
        if not pd.api.types.is_numeric_dtype(points[arg]):
            raise TypeError(f"{arg_name} column must be numeric dtype.")
        if points[arg].isnull().any():
            raise ValueError(f"{arg_name} column must be complete.")
        arg = points[arg].to_numpy()
    elif isinstance(arg, (int, float)):
        arg = np.full(len(points), arg)
    else:
        raise ValueError(
            f"{arg_name} must correspond to a GeoDataFrame column or be int or float.",
        )
    return arg


def create_array(
    bounds: np.ndarray,
    radius: int | float,
    resolution: int | float,
) -> tuple[np.ndarray, list[int | float]]:
    """Generate an ndarray of shape derived from bounding coordinates and spatial
    resolution and calculate the radius- and shape-adjusted bounds thereof.

    Parameters
    ----------
    bounds : numpy.ndarray
        Bounding coordinates to adjust.
    radius : int | float
        Radius with which to adjust coordinates.
    resolution : int | float
        Spatial resolution for the output array.

    Returns
    -------
    array : numpy.ndarray
        Array of shape determined using bounding coordinates and resolution.
    list[int | float]
        List of minx, miny, maxx, and maxy bounding coordinates for the array.
    """
    minx, miny, maxx, maxy = bounds
    minx -= radius
    maxy += radius
    width = round((maxx + radius - minx) / resolution)
    height = round((maxy - (miny - radius)) / resolution)
    maxx = minx + width * resolution
    miny = maxy - height * resolution
    array = np.full((height, width), 0.0)
    return array, [minx, miny, maxx, maxy]


def get_points(
    points: gpd.GeoDataFrame | gpd.GeoSeries,
    minx: int | float,
    maxy: int | float,
    radius: np.ndarray,
    weight: np.ndarray,
    resolution: int | float,
) -> np.ndarray:
    """Generate an ndarray of point X and Y array coordinates, search window radii, and
    weights.

    Parameters
    ----------
    points : geopandas.GeoDataFrame | geopandas.GeoSeries
        A GeoDataFrame or GeoSeries of point geometries from which to derive array
        coordinates.
    minx : int | float
        Minimum X bounding coordinate of the array for which coordinates are generated.
    maxy : int | float
        Maximum Y bounding coordinate of the array for which coordinates are generated.
    radius : numpy.ndarray
        Radii with which to calculate a point's search window radius.
    weight : numpy.ndarray
        Values with which a point's KDE values will be weighted.
    resolution : int | float
        Spatial resolution of the array to which KDE values will be added.

    Returns
    -------
    array_points : numpy.ndarray
        Points comprising array X and Y coordinates, search window radius, and weight.
    """
    x_idx = (points.geometry.x - minx) / resolution
    y_idx = (maxy - points.geometry.y) / resolution
    window = radius / resolution
    array_points = np.column_stack((x_idx, y_idx, window, weight))
    return array_points


def calculate_kde(
    points: np.ndarray,
    array: np.ndarray,
    kernel: str,
    scale: bool,
) -> None:
    """Iterate over elements of an ndarray comprising point properties.

    Parameters
    ----------
    points : numpy.ndarray
        Points comprising array X and Y coordinates, search window radius, and weight
        for which KDE will be calculated.
    array : numpy.ndarray
        Array to which KDE values will be added.
    kernel : str
        Kernel with which to perform KDE.
    scale : bool
        Whether to calculate raw or scaled KDE values.
    """
    kernel_vfuncs = {
        "epanechnikov": epanechnikov_scaled if scale else epanechnikov_raw,
        "quartic": quartic_scaled if scale else quartic_raw,
        "triweight": triweight_scaled if scale else triweight_raw,
    }
    # challenge to vectorise as needs to operate on >1 array element
    for point in points:
        add_point_kde(point, array, kernel_vfuncs[kernel])


def add_point_kde(
    point: np.ndarray,
    array: np.ndarray,
    kernel_vfunc: np.vectorize,
) -> None:
    """Perform KDE for a given point, adding the result to an array.

    Parameters
    ----------
    point : numpy.ndarray
        Properties of the point for which KDE will be calculated, including array X and
        Y coordinates, search window radius, and weight.
    array : numpy.ndarray
        Array to which KDE values will be added.
    kernel_vfunc : numpy.vectorize
        Vectorised kernel function with which to perform KDE.
    """
    x, y, window, weight = point
    minx = round(x - window)
    miny = round(y - window)
    maxx = round(x + window)
    maxy = round(y + window)
    y_idx, x_idx = np.ogrid[miny + 0.5 : maxy + 0.5, minx + 0.5 : maxx + 0.5]
    dist_array = np.sqrt(pow(x_idx - x, 2) + pow(y_idx - y, 2))
    kde_array = kernel_vfunc(dist_array, window, weight)
    array[miny:maxy, minx:maxx] += kde_array



import geopandas as gpd
import numpy as np

from geokde._kernels import VALID_KERNELS
from geokde._utils import (
    calculate_kde,
    create_array,
    get_points,
    validate_transform,
)


def kde(
    points: gpd.GeoDataFrame | gpd.GeoSeries,
    radius: int | float | str,
    resolution: int | float,
    kernel: str = "quartic",
    weight: int | float | str = 1.0,
    scale: bool = False,
) -> tuple[np.ndarray, list[int | float]]:
    """Estimate raw or scaled kernel density with a given radius (or radii), resolution,
    kernel, and weight(s) from point geometries in a given GeoDataFrame or GeoSeries.

    Parameters
    ----------
    points : geopandas.GeoDataFrame | geopandas.GeoSeries
        A GeoDataFrame or GeoSeries of point geometries.
    radius : int | float | str
        Value(s) to use as the search radius.Can be a str, int, or float corresponding
        to a column name in the GeoDataFrame (used preferentially if present), or a
        single int or float value. Should be in the same units as the `points`
        GeoDataFrame's CRS.
    resolution : int | float
        Spatial resolution of the output KDE array. Should be in the same units as the
        `points` GeoDataFrame's CRS.
    kernel : str, default = "quartic"
        Kernel type to use in density estimation. Must be one of "epanechnikov",
        "quartic", or "triweight". Default is "quartic".
    weight : int | float | str, default = 1.0
        Value(s) to use as weights. Can be a str, int, or float corresponding to a
        column name in the GeoDataFrame (used preferentially if present), or a single
        int or float value.
    scale : bool, default = False
        Whether to calculate raw or scaled KDE values, defaults to False (raw).

    Returns
    -------
    array : numpy.ndarray
        Array of KDE values.
    bounds : list[int | float]
         The array's bounding coordinates in minx, miny, maxx, maxy format.

    Raises
    ------
    ValueError
        If the specified kernel is invalid or resolution is greater than the maximum
        specified radius.
    TypeError
        If any geometries are not a point or scale is not boolean.

    Examples
    --------
    Writing results with rasterio:

    >>> gdf = geopandas.read_file("vector_points.geojson")
    >>> kde_array, array_bounds = geokde.kde(gdf, 1, 0.1)
    >>> transform = rasterio.transform.from_bounds(
        *array_bounds,
        kde_array.shape[1],
        kde_array.shape[0],
    )

    >>> with rasterio.open(
        fp="raster.tif",
        mode="w",
        driver="GTiff",
        width=kde_array.shape[1],
        height=kde_array.shape[0],
        count=1,
        crs=gdf.crs,
        transform=transform,
        dtype=kde_array.dtype,
        nodata=0.0,
    ) as dst:
        dst.write(kde_array, 1)
    """
    if not all(points.geometry.geom_type == "Point"):
        raise TypeError("all geometries must be points.")
    if kernel not in VALID_KERNELS:
        raise ValueError(f"kernel must be one of {VALID_KERNELS}, not: {kernel}")
    if not isinstance(scale, bool):
        raise TypeError(f"scale must be bool, not: {type(scale)}")
    radius_arr = validate_transform(radius, "radius", points)
    weight_arr = validate_transform(weight, "weight", points)
    if resolution > radius_arr.max():
        raise ValueError("resolution must be less than radius.")

    array, bounds = create_array(points.total_bounds, radius_arr.max(), resolution)
    points = get_points(
        points,
        bounds[0],
        bounds[3],
        radius_arr,
        weight_arr,
        resolution,
    )
    calculate_kde(points, array, kernel, scale)
    return array, bounds



"""Kernels for density estimation.

All kernels accept the following parameters:

Parameters
----------
distance : int | float
    Euclidean distance value for a given point from the reference point.
radius : int | float
    Search window radius for the reference point.
weight : int | float
    Value with which the KDE value for a point will be weighted.

And provide the following return:

Returns
-------
value : float
    Specified kernel's density estimation value.
"""

import numpy as np


VALID_KERNELS = [
    "epanechnikov",
    "quartic",
    "triweight",
]


@np.vectorize
def quartic_raw(
    distance: int | float,
    radius: int | float,
    weight: int | float,
) -> float:
    """Raw Quartic kernel."""
    if distance < radius:
        value = weight * pow(1 - pow(distance / radius, 2), 2)
    else:
        value = 0.0
    return value


@np.vectorize
def quartic_scaled(
    distance: int | float,
    radius: int | float,
    weight: int | float,
) -> float:
    """Scaled Quartic kernel."""
    if distance < radius:
        norm_const = 116 / (5 * np.pi * pow(radius, 2))
        value = weight * (
            norm_const * (15 / 16) * pow(1 - pow(distance / radius, 2), 2)
        )
    else:
        value = 0.0
    return value


@np.vectorize
def epanechnikov_raw(
    distance: int | float,
    radius: int | float,
    weight: int | float,
) -> float:
    """Raw Epanechnikov kernel."""
    if distance < radius:
        value = weight * (1 - pow(distance / radius, 2))
    else:
        value = 0.0
    return value


@np.vectorize
def epanechnikov_scaled(
    distance: int | float,
    radius: int | float,
    weight: int | float,
) -> float:
    """Scaled Epanechnikov kernel."""
    if distance < radius:
        norm_const = 8 / (3 * np.pi * pow(radius, 2))
        value = weight * (norm_const * (3 / 4) * (1 - pow(distance / radius, 2)))
    else:
        value = 0.0
    return value


@np.vectorize
def triweight_raw(
    distance: int | float,
    radius: int | float,
    weight: int | float,
) -> float:
    """Raw triweight kernel."""
    if distance < radius:
        value = weight * pow(1 - pow(distance / radius, 2), 3)
    else:
        value = 0.0
    return value


@np.vectorize
def triweight_scaled(
    distance: int | float,
    radius: int | float,
    weight: int | float,
) -> float:
    """Scaled triweight kernel."""
    if distance < radius:
        norm_const = 128 / (35 * np.pi * pow(radius, 2))
        value = weight * (
            norm_const * (35 / 32) * pow(1 - pow(distance / radius, 2), 3)
        )
    else:
        value = 0.0
    return value
